module("L_Sonos1", package.seeall)

local http = require("socket.http")

-- 5 Second timeout
http.TIMEOUT = 5

local UPNP_REQUEST = [[<?xml version="1.0" encoding="utf-8"?>
<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
<s:Body>
<u:%s xmlns:u="%s">%s</u:%s>
</s:Body>
</s:Envelope>]]

UPNP_AVTRANSPORT_SERVICE = 'urn:schemas-upnp-org:service:AVTransport:1'
UPNP_RENDERING_CONTROL_SERVICE = 'urn:schemas-upnp-org:service:RenderingControl:1'
UPNP_DEVICE_PROPERTIES_SERVICE = 'urn:schemas-upnp-org:service:DeviceProperties:1'

UPNP_AVTRANSPORT_SID = 'urn:upnp-org:serviceId:AVTransport'
UPNP_RENDERING_CONTROL_SID = 'urn:upnp-org:serviceId:RenderingControl'
UPNP_DEVICE_PROPERTIES_SID = 'urn:upnp-org:serviceId:DeviceProperties'

local DIDL_FORMAT=[[<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">
<item id="%s" parentID="%s" restricted="true">
<res protocolInfo="x-file-cifs:*:audio/mpeg:*">%s</res>
<r:streamContent>%s</r:streamContent>
<upnp:albumArtURI>%s</upnp:albumArtURI>
<dc:title>%s</dc:title>
<upnp:class>%s</upnp:class>
<dc:creator>%s</dc:creator>
<upnp:album>%s</upnp:album>
<upnp:originalTrackNumber>%s</upnp:originalTrackNumber>
<r:albumArtist>%s</r:albumArtist>
</item>
</DIDL-Lite>
]]

local log = print
local contentType

function initialize(logger, ct)
  log = logger
  contentType = ct or "application/x-www-form-urlencoded"
end

local function decode(val)
      return val:gsub("&amp;", "&")
                :gsub("&lt;", "<")
                :gsub("&gt;", ">")
                :gsub("&quot;", '"')
end

local function encode(val)
      return val:gsub("&", "&amp;")
                :gsub("<", "&lt;")
                :gsub(">", "&gt;")
                :gsub('"', "&quot;")
end

function createDIDL(resURI, creator, album, title, artist, streamContent, albumArtURI, trackNumber, class, id, parentID)
  return DIDL_FORMAT:format(encode(id or 1),
                            encode(parentID or -1),
                            encode(resURI or ""),
                            encode(streamContent or ""),
                            encode(albumArtURI or ""),
                            encode(title or "No title"),
                            encode(class or "object.item.audioItem.musicTrack"),
                            encode(creator or ""),
                            encode(album or ""),
                            encode(trackNumber or 1),
                            encode(artist or "No artist"))
end

-- event url, user-agent, callback, nt, TimeOut
function UPnP_subscribe(ipAddress, ipPort, serverURL, callbackURL)
  --
  -- Execute the subscription
  --
  local status, statusMsg = http.request{
    url = serverURL:format(ipAddress, ipPort),
    method = "SUBSCRIBE",
    headers = {["HOST"] = string.format("%s:%s", ipAddress, ipPort),
               ["CALLBACK"] = string.format("<%s>", callbackURL),
               ["NT"] = "upnp:event"}
  }

  return status
end  
  
  
function UPnP_request(ipAddress, ipPort, serverURL, action, servicetype, args)
  local function table2XML(value)
    --
    -- Encode all the usual suspects
    --
    local function encode(val)
      return val:gsub("&", "&amp;")
                :gsub("<", "&lt;")
                :gsub(">", "&gt;")
                :gsub('"', "&quot;")
    end

    local result = ""

    --
    -- Convert all the Number, Boolean and Table objects, and escape all the string
    -- values in the XML output stream
    --
    for e, v in pairs(value) do
      if (v == nil) then
          result = result .. string.format("<%s />", e)
      elseif (type(v) == "table") then
          result = result .. table2XML(v)
      elseif (type(v) == "number") then
          result = result .. string.format("<%s>%.0f</%s>", e, v, e)
      elseif (type(v) == "boolean") then
          result = result .. string.format("<%s>%s</%s>", e, (v and "1" or "0"), e)
      else
          result = result .. string.format("<%s>%s</%s>", e, encode(v), e)
      end
    end

    return result
  end

  local postBody = string.format(UPNP_REQUEST, action, servicetype, table2XML(args), action)
  serverURL = serverURL:format(ipAddress, ipPort)
  log(string.format("UPnP_request: url=[%s], body=[%s]", serverURL, postBody))

  --
  -- Execute the resulting URL, and collect the results as a Table
  --
  local resultTable = {}
  local status, statusMsg = http.request{
    url = serverURL,
    sink = ltn12.sink.table(resultTable),
    method = "POST",
    headers = {["Accept"] = "*/*",
               ["SOAPAction"] = servicetype .. "#" .. action,
               ["Connection"] = "close",
               ["Content-Length"] = postBody:len(),
               ["Content-Type"] = contentType},
    source = ltn12.source.string(postBody),
  }

  --
  -- Flatten the resultTable into a regular string
  --
  local data = ""
  for i, v in ipairs(resultTable) do
    data = data .. v
  end
  
  log(string.format("UPnP_request: status=%s statusMsg=%s result=[%s]", status or "no status",
                                              statusMsg or "no message",
                                              data or "no result"))

  if (status == nil) then
    --
    -- Handle TIMEOUT
    --
    return false, statusMsg or "An error occurred during the UPnP call"
  end
  
  if (tostring(statusMsg) == "200") then
    --
    -- Handle SUCCESS
    --
    if (data == nil or data == "") then
      return true, ""
    else
      local pattern = string.format('.*<s:Body><u:%sResponse xmlns:u="urn:(.*)">(.*)</u:%sResponse></s:Body></s:Envelope>.*', action, action)

      -- TODO: Handle UPnP Error responses
      -- TODO: Handle Non-Scalar results
      -- luup.log("Before gmatch: " .. data)
      local sid, value
      for serviceId, result in data:gmatch(pattern) do
        -- luup.log(serviceId .. " --> " .. result)
        sid, value = serviceId, result
      end

      -- commented out and returns a simple string instead. Due to multiple tags (not nested though)
      --      [[local dataTable = {}
      --      for tagBegin, value, tagEnd in value:gmatch("<(.*)>(.*)</(.*)>") do
      --        print(tagBegin .. " -> " .. value .. " <- " .. tagEnd)
      --        dataTable[tagBegin] = value
      --      end

      --     return dataTable]]
      return true, decode(value)
    end
  else
    --
    -- UNKNOWN ERROR
    --
    assert("Unhandled Response statusMsg=" .. statusMsg)
  end
end -- function UPnP_request

